#!/usr/bin/env python

"""Flasker command line tool.

There are currently 4 commands available via the flasker command tool, all
detailed below.

All commands accept an optional argument ``-c, --conf``
to indicate the path of the configuration file to use. If none is specified
flasker will search in the current directory for possible matches. If a single
file ``.cfg`` file is found it will use it.

"""

from argparse import ArgumentParser, REMAINDER
from code import interact
from functools import wraps
from os import listdir
from os.path import splitext
from re import findall, match, sub

from flasker.project import current_project as pj, Project, ProjectImportError


def _project_context(handler):
  """Create the project context.

  Subparser handlers require the project to be created before
  returning, this decorator handles this.

  """
  @wraps(handler)
  def wrapper(*args, **kwargs):
    parsed_args = args[0]
    try:
      conf_file = parsed_args.conf
      if not conf_file:
        conf_files = [fn for fn in listdir('.') if splitext(fn)[1] == '.cfg']
        if len(conf_files) == 0:
          print (
            'No configuration file found in current directory. '
            'Please enter a path to one with the -c option.'
          )
          return
        elif len(conf_files) > 1:
          print (
            'Several configuration files found in current directory: %s. '
            'Please disambiguate with the -c option.'
          ) % ', '.join(conf_files)
          return
        else:
          conf_file = conf_files[0]
      pj = Project(conf_file)
    except ProjectImportError as e:
      print e
      return
    else:
      handler(*args, **kwargs)
  return wrapper


# Parsers

# Main parser

parser = ArgumentParser('flasker')

parser.add_argument('-c', '--conf',
  dest='conf',
  default='',
  help='path to configuration file'
)
subparsers = parser.add_subparsers(
  title='available commands',
  dest='command',
)


# Server

server_parser = subparsers.add_parser('server', help='start server')

server_parser.add_argument('-r', '--restrict',
  action='store_true',
  help='disallow remote server connections'
)
server_parser.add_argument('-p', '--port',
  default=5000,
  type=int,
  help='listen on port [%(default)s]'
)
server_parser.add_argument('-d', '--debug',
  action='store_true',
  help='run in debug mode (autoreload and debugging)'
)

@_project_context
def server_handler(parsed_args):
  """Start a Werkzeug server for the Flask application::
  
    flasker server ...

  The following options are available:

  * ``-r, --restrict`` to disallow remove server connections.
  * ``-p, --port`` to set the port on which to run the server (default to 
    ``5000``).
  * ``-d, --debug`` to run in debug mode (enables autoreloading and in-browser
    debugging).
  
  """
  host = '127.0.0.1' if parsed_args.restrict else '0.0.0.0'
  pj.flask.run(host=host, port=parsed_args.port, debug=parsed_args.debug)

server_parser.set_defaults(handler=server_handler)

# Shell

shell_parser = subparsers.add_parser('shell', help='start shell')

@_project_context
def shell_handler(parsed_args):
  """Start a shell in the context of the project::

    flasker shell ...

  The following global variables will be available:

  * ``pj``, an alias for the ``current_project``

  Flasker will use IPython if it is available.
  
  """
  context = {'pj': pj}
  try:
    import IPython
  except ImportError:
    interact(local=context)
  else:
    sh = IPython.frontend.terminal.embed.InteractiveShellEmbed()
    sh(local_ns=context)

shell_parser.set_defaults(handler=shell_handler)

# Worker

worker_parser = subparsers.add_parser('worker', help='start worker')

worker_parser.add_argument('domain',
  help='worker domain name'
)
worker_parser.add_argument('-o', '--only-direct',
  action='store_true',
  help='only listen to direct queue'
)
worker_parser.add_argument('-v', '--verbose-help',
  action='store_true',
  help='show full help from celery worker'
)
worker_parser.add_argument('-r', '--raw',
  nargs=REMAINDER,
  help='raw options to pass through',
)

@_project_context
def worker_handler(parsed_args):
  """Starts a celery worker::

    flasker worker ...

  The following options are available:

  * ``-o, --only-direct`` to have the worker only listen to its direct queue
    (this option requires the CELERY_WORKER_DIRECT to be set to ``True``).
  * ``-v, --verbose-help`` to show the worker help.
  * ``-r, --raw`` to pass arguments to the worker (any arguments after this
    option will be passed through).

  If no hostname is provided, one will be generated automatically using the
  project domain and subdomain and current worker count. For example, the first
  two workers started for project ``my_project`` and configuration ``default``
  will have respective hostnames:

  * w1.default.my_project
  * w2.default.my_project

  """
  domain = parsed_args.domain
  pj_worker_names = [d.keys()[0] for d in pj.celery.control.ping()]
  worker_pattern = r'w(\d+)\.%s' % (domain, )
  worker_numbers = [
    findall(worker_pattern, worker_name) or ['0']
    for worker_name in pj_worker_names
  ]
  wkn = min(
    set(range(1, len(worker_numbers) + 2)) -
    set([int(n[0]) for n in worker_numbers] or [len(worker_numbers) + 2]) 
  )
  if parsed_args.verbose_help:
    pj.celery.worker_main(['worker', '-h'])
  else:
    hostname = 'w%s.%s' % (wkn, domain)
    options = ['worker', '--hostname=%s' % hostname]
    if parsed_args.only_direct:
      options.append('--queues=%s.dq' % hostname)
    if parsed_args.raw:
      options.extend(parsed_args.raw)
    pj.celery.worker_main(options)

worker_parser.set_defaults(handler=worker_handler)

# Flower

flower_parser = subparsers.add_parser('flower', help='start flower')

flower_parser.add_argument('-p', '--port',
  default=5555,
  type=int,
  help='listen on port [%(default)s]'
)
flower_parser.add_argument('-v', '--verbose-help',
  action='store_true',
  help='show full help from celery flower'
)
flower_parser.add_argument('-r', '--raw',
  help='raw options to pass through',
  nargs=REMAINDER
)

@_project_context
def flower_handler(parsed_args):
  """Start flower worker manager::
    
    flasker flower ...

  The following arguments are available:

  * ``-p, --port`` to set the port to run flower on (defaults to ``5555``).
  * ``-v, --verbose-help`` to show the flower help.
  * ``-r, --raw`` to pass arguments to flower (any arguments after this option
    will be passed through).
  
  """
  if parsed_args.verbose_help:
    pj.celery.start(['celery', 'flower', '--help'])
  else:
    options = ['celery', 'flower', '--port=%s' % parsed_args.port]
    if parsed_args.raw:
      options.extend(parsed_args.raw)
    pj.celery.start(options)

flower_parser.set_defaults(handler=flower_handler)

# END of parsers

def main():
  parsed_args = parser.parse_args()
  parsed_args.handler(parsed_args)

if __name__ == '__main__':
  main()



#!/usr/bin/env python

"""ORM Extension

This extension provides a customized SQLAlchemy model base and query.

Setup is straightforward:

.. code:: python

  from flasker import current_project as pj
  from flasker.ext import ORM

  orm = ORM(pj)

  Model = orm.Model                 # the customized base
  relationship = orm.relationship   # the customized relationship
  backref = orm.backref             # the associated backref

Models can now be created by subclassing ``orm.Model`` as follows:

.. code:: python

  from sqlalchemy import Column, ForeignKey, Integer, String

  class House(Model):

    id = Column(Integer, primary_key=True)
    address = Column(String(128))

  class Cat(Model):
      
    id = Column(Integer, primary_key=True)
    name = Column(String(64))
    house_id = Column(ForeignKey('houses.id'))

    house = relationship('House', backref=backref('cats', lazy='dynamic'))

Note that tablenames are automatically generated by default. For an
exhaustive list of all the properties and methods provided by ``orm.Model``
please refer to the documentation for :class:`flasker.util._sqlalchemy.Model`.

Models can be queried in several ways:

.. code:: python

  # the two following queries are equivalent
  query = pj.session.query(Cat)
  query = Cat.q

Both queries above are instances of :class:`flasker.ext.orm.Query`, which are
customized ``sqlalchemy.orm.Query`` objects (cf. below for the list of
available methods). If relationships (and backrefs) are defined using the
``orm.relationship`` and ``orm.backref`` functions, appender queries will
also return custom queries:

.. code:: python

  house = House.q.first()
  relationship_query = house.cats   # instance of flasker.ext.orm.Query


"""

from functools import partial
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import (backref as _backref, class_mapper,
  relationship as _relationship)
from sqlalchemy.orm.exc import UnmappedClassError

from ..util.sqlalchemy import Model, Query


class ORM(object):

  """The main ORM object.

  :param project: the project against which the extension will be registered
  :type project: flasker.project.Project
  :param create_all: whether or not to automatically create tables for the
    models defined (``True`` by default). Tables will only be created for
    models which do not have one already.
  :type create_all: bool

  """

  def __init__(self, project, create_all=True):

    project.conf['SESSION']['QUERY_CLS'] = Query

    self.Model = declarative_base(cls=Model)
    self.Model.q = _QueryProperty(project)
    self.Model.t = _TableProperty(project)

    self.backref = partial(_backref, query_class=Query)
    self.relationship = partial(_relationship, query_class=Query)

    @project.run_after_module_imports
    def orm_after_imports(project):
      if create_all:
        self.Model.metadata.create_all(
          project.session.get_bind(),
          checkfirst=True
        )

    project.logger.debug('orm extension initialized')


class _QueryProperty(object):

  """To make queries accessible directly on model classes."""

  def __init__(self, project):
    self.project = project

  def __get__(self, obj, cls):
    try:
      mapper = class_mapper(cls)
      if mapper:
        return Query(mapper, session=self.project.session())
    except UnmappedClassError:
      return None


class _TableProperty(object):

  """Bound table for faster batch executes."""

  def __init__(self, project):
    self.project = project

  def __get__(self, obj, cls):
    try:
      mapper = class_mapper(cls)
      if mapper:
        table = mapper.mapped_table
        table.metadata.bind = self.project.session.connection()
        return table
    except UnmappedClassError:
      return None




#!/usr/bin/env python

"""General helpers."""

from __future__ import absolute_import

from distutils.dir_util import copy_tree
from flask import request
from flask.views import View as _View
from os import mkdir
from os.path import dirname, exists, join


class _ViewMeta(type):

  """To register classes with an app or blueprint on definition."""

  http_methods = ['get', 'post', 'put', 'delete']

  def __new__(cls, name, bases, dct):
    rv = super(_ViewMeta, cls).__new__(cls, name, bases, dct)

    if rv.rules:

      if not rv.__app__:
        raise ValueError('%s is not bound to an app' % (rv, ))

      if rv.endpoint is None:
        rv.endpoint = uncamelcase(rv.__name__)

      if rv.methods is None:
        methods = set(key.upper() for key in dct if key in cls.http_methods)
        rv.methods = sorted(methods or [])
      
      rv.register_view(rv.__app__)

    return rv


class View(_View):

  """Base view class.

  Not to be used directly, should be subclassed or called via the function
  :func:`make_view`.

  If bound to an app, will automatically register itself for the rules defined
  in the rules property.
  
  """

  __metaclass__ = _ViewMeta
  __app__ = None

  endpoint = None
  methods = None
  rules = None

  @classmethod
  def register_view(cls, app):
    """Attach view to app or blueprint."""
    view = cls.as_view(cls.endpoint)

    all_methods = set(cls.methods)
    if cls.rules is None:
      raise ValueError('No rules found for %r' % (cls, ))
    for rule, methods in cls.rules.items():
      rule_methods = set(methods) & all_methods
      if rule_methods:
        app.add_url_rule(rule=rule, view_func=view, methods=rule_methods)
      
  def dispatch_request(self, **kwargs):
    """Dispatches requests to the corresponding method name.
    
    Similar to the ``flask.views.MethodView`` implementation: GET requests
    are passed to :meth:`get`, POST to :meth:`post`, etc.
    
    """
    meth = getattr(self, request.method.lower(), None)
    if meth is None and request.method == 'HEAD':
        meth = getattr(self, 'get', None)
    return meth(**kwargs)


def make_view(app, view_class=View, view_name='View', **kwargs):
  """Return base view class bound to app.

  :param app: the app (or blueprint) to be bound to
  :type app: Flask app or blueprint
  :param view_class: base view class
  :type view_class: flasker.util.View
  :rtype: flasker.util.View

  Any keyword arguments will be added to the class' dictionary.

  """
  kwargs.update({'__app__': app})
  return type(view_name, (view_class, ), kwargs)

def generate_app(foldername='app'):
  """Create bootstrap app."""
  src = dirname(__file__)
  if exists(foldername):
    print 'That folder already seems to exist. App creation skipped.'
  else:
    mkdir(foldername)
    copy_tree(join(src, 'data', 'app'), foldername)
    print 'Bootstrap app folder created in %s!' % (foldername, )




#!/usr/bin/env python

"""SQAlchemy helpers."""

from __future__ import absolute_import

from json import dumps, loads
from random import randint
from sqlalchemy import Column, func
from sqlalchemy.ext.associationproxy import AssociationProxy
from sqlalchemy.ext.declarative import declared_attr 
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import class_mapper, Query as _Query
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.types import TypeDecorator, UnicodeText
from sqlalchemy.orm.properties import ColumnProperty, RelationshipProperty

try:
  from pandas import DataFrame
except ImportError:
  pass

from .helpers import uncamelcase, to_json
from .mixins import Cacheable, Loggable


# Base model

class Model(Cacheable, Loggable):

  """The custom SQLAlchemy base.

  Along with the methods described below, the following conveniences are
  provided:

  * Automatic table naming (to the model's class name uncamelcased with an
    extra s appended for good measure). To disable this behavior, simply
    override the ``__tablename__`` argument (setting it to ``None`` for
    single table inheritance).

  * Default implementation of ``__repr__`` with model class and primary keys

  * Caching (inherited from :class:`flasker.util.Cacheable`). The cache is
    persistent by default (``_cache`` is actually a
    :class:`flasker.util.JSONEncodedDict` column).

  * Logging (inherited from :class:`flasker.util.Loggable`)

  """

  @classmethod
  def __declare_last__(cls):
    """Creates the ``__json__`` attribute.
    
    Varnames that get JSONified. Doesn't emit any additional queries!

    TODO: use _get_columns and other methods to generate thist list.

    """
    cls.__json__ = list(
      varname
      for varname in dir(cls)
      if not varname.startswith('_')  # don't show private properties
      if (
        isinstance(getattr(cls, varname), property) 
      ) or (
        isinstance(getattr(cls, varname), InstrumentedAttribute) and
        isinstance(getattr(cls, varname).property, ColumnProperty)
      ) or (
        isinstance(getattr(cls, varname), InstrumentedAttribute) and
        isinstance(getattr(cls, varname).property, RelationshipProperty) and
        getattr(cls, varname).property.lazy in [False, 'joined', 'immediate']
      ) or (
        isinstance(getattr(cls, varname), AssociationProxy) and
        getattr(
          cls, getattr(cls, varname).target_collection
        ).property.lazy in [False, 'joined', 'immediate']
      )
    )

  @classmethod
  def _get_columns(cls, show_private=False):
    return {
      c.key: c
      for c in class_mapper(cls).columns
      if show_private or not c.key.startswith('_')
    }

  @classmethod
  def _get_related_models(cls, show_private=False):
    return {
      k: v.mapper.class_
      for k, v in cls._get_relationships(show_private).items()
    }

  @classmethod
  def _get_relationships(cls, show_private=False, lazy=None, uselist=None):
    return {
      rel.key: rel
      for rel in class_mapper(cls).relationships.values()
      if show_private or not rel.key.startswith('_')
      if lazy is None or rel.lazy in lazy
      if uselist is None or rel.uselist == uselist
    }

  @classmethod
  def _get_association_proxies(cls, show_private=False):
    return {
      varname: getattr(cls, varname)
      for varname in dir(cls)
      if isinstance(getattr(cls, varname), AssociationProxy)
      if show_private or not varname.startswith('_')
    }

  @classmethod
  def retrieve(cls, from_key=False, if_not_found='pass', **kwargs):
    """Given constructor arguments will return a match or create one.

    :param if_not_found: whether or not to create and flush the model if 
      created (this can be used to generate its ``id``). Acceptable values:
      'flush' (create and flush), 'create' (only create), 'pass' (do nothing).
    :type if_not_found: str
    :param from_key: instead of issuing a filter on kwargs, this will issue
      a get query by id using this parameter. Note that in this case, any other
      keyword arguments will only be used if a new instance is created.
    :type from_key: bool
    :param kwargs: constructor arguments
    :rtype: tuple

    This method returns a tuple ``(model, flag)`` where ``model`` is of the
    corresponding class and ``flag`` is ``True`` if the model was just created
    and ``False`` otherwise.

    """
    if from_key:
      model_primary_key = tuple(
        kwargs[k.name]
        for k in class_mapper(cls).primary_key
      )
      instance = cls.q.get(model_primary_key)
    else:
      instance = cls.q.filter_by(**kwargs).first()
    if if_not_found == 'pass':
      return instance
    else:
      if instance:
        return instance, False
      else:
        instance = cls(**kwargs)
        if if_not_found == 'flush':
          instance.flush()
      return instance, True

  def __repr__(self):
    primary_keys = ', '.join(
      '%s=%r' % (k, getattr(self, k))
      for k, v in self.get_primary_key().items()
    )
    return '<%s (%s)>' % (self.__class__.__name__, primary_keys)

  @declared_attr
  def __tablename__(cls):
    """Automatically create the table name."""
    return '%ss' % uncamelcase(cls.__name__)

  @declared_attr
  def _cache(cls):
    """Automatically create the table name."""
    return Column(JSONEncodedDict)

  def flush(self, merge=False):
    """Add the model to the session and flush.
    
    :param merge: if ``True``, will merge instead of add.
    :type merge: bool
    
    """
    session = self.q.session
    if merge:
      session.merge(self)
    else:
      session.add(self)
    session.flush([self])

  def get_primary_key(self, as_tuple=False):
    """Returns a dictionary of primary keys for the given model.

    :param as_tuple: if set to ``True``, this method will return a tuple with
      the model's primary key values. Otherwise a dictionary is returned.
    :type as_tuple: bool
    :rtype: dict, tuple

    """
    if as_tuple:
      return tuple(
        getattr(self, k.name)
        for k in class_mapper(self.__class__).primary_key
      )
    else:
      return dict(
        (k.name, getattr(self, k.name))
        for k in class_mapper(self.__class__).primary_key
      )

  def to_json(self, depth=1):
    """Serializes the model into a dictionary.

    :param depth:
    :type depth: int
    :rtype: dict

    The following attributes are included in the returned JSON:

    * all non private columns
    * all non private properties
    * all non private relationships which have their ``lazy`` attribute set to
      one of ``False, 'joined', 'immediate'``

    A consequence of this is that this method will never issue extra queries
    to populate the JSON. Furthermore, all the attribute names to be
    included are computed at class declaration so this method is very fast.

    .. note::

      To change which attributes are included in the dictionary, you can 
      override the ``__json__`` attribute.

    """
    if depth <= 0:
      return self.get_primary_key()
    rv = {}
    for varname in self.__json__:
      try:
        rv[varname] = to_json(getattr(self, varname), depth - 1)
      except ValueError as e:
        rv[varname] = e.message
    return rv


# Query helpers

def query_to_models(query):
  """Returns the model classes associated with a query.
  
  :param query: the query to be executed
  :type query: sqlalchemy.orm.query.Query
  :rtype: list

  """
  if hasattr(query, 'attr'):
    # this is a subquery
    return [query.attr.target_mapper]
  else:
    # this is a main query
    return [
      d['expr'].class_
      for d in query.column_descriptions
      if isinstance(d['expr'], Mapper)
    ]

def query_to_dataframe(query, connection=None, exclude=None, index=None,
                       columns=None, coerce_float=False):
  """Load a Pandas dataframe from an SQLAlchemy query.

  :param query: the query to be executed
  :type query: sqlalchemy.orm.query.Query
  :param connection: the connection to use to execute the query. By default
    the method will create a new connection using the session's bound engine
    and properly close it afterwards.
  :type connection: sqlalchemy.engine.base.Connection
  :param exclude: a list of column names to exclude from the dataframe
  :type exclude: list
  :param index: the column to use as index
  :type index: str
  :param names: a list of column names. If unspecified, the method will use
    the table's keys from the query's metadata. If the passed data do not have
    named associated with them, this argument provides names for the columns.
    Otherwise this argument indicates the order of the columns in the result
    (any names not found in the data will become all-NA columns)
  :type names: list
  :param coerce_float: Attempt to convert values to non-string, non-numeric
    objects (like decimal.Decimal) to floating point.
  :type coerce_float: bool
  :rtype: pandas.DataFrame
  
  """
  connection = connection or query.session.get_bind()
  exclude = exclude or []
  result = connection.execute(query.statement)
  columns = columns or result.keys()
  dataframe = DataFrame.from_records(
    result.fetchall(),
    columns=columns,
    exclude=exclude,
    index=index,
    coerce_float=coerce_float,
  )
  result.close()
  return dataframe

def query_to_records(query, connection=None):
  """Raw execute of the query into a generator.

  :param query: the query to be executed
  :type query: sqlalchemy.orm.query.Query
  :param connection: the connection to use to execute the query. By default
    the method will create a new connection using the session's bound engine
    and properly close it afterwards.
  :type connection: sqlalchemy.engine.base.Connection
  :rtype: generator

  About 5 times faster than loading the objects. Useful if only interested in
  raw columns of the model::

    # for ~300k models
    In [1]: %time [m.id for s in Model.q]
    CPU times: user 48.22 s, sys: 2.62 s, total: 50.84 s
    Wall time: 52.19 s
    In [2]: %time [m['id'] for s in query_to_records(Model.q)]
    CPU times: user 9.12 s, sys: 0.20 s, total: 9.32 s
    Wall time: 10.32 s
  
  """
  connection = connection or query.session.get_bind()
  result = connection.execute(query.statement)
  keys = result.keys()
  for record in result:
    yield {k:v for k, v in zip(keys, record)}
  result.close()


class Query(_Query):

  """Base query class.

  All queries and relationships/backrefs defined using this extension will
  return an instance of this class.

  """

  def get_or_404(self, model_id):
    """Like get but aborts with 404 if not found.

    :param model_id: the model's primary key
    :type model_id: varies
    :rtype: model or HTTPError

    This method is from Flask-SQLAlchemy.
    
    """
    rv = self.get(model_id)
    if rv is None:
      abort(404)
    return rv

  def first_or_404(self):
    """Like first but aborts with 404 if not found.
    
    :rtype: model or HTTPError

    This method is from Flask-SQLAlchemy.
    
    """
    rv = self.first()
    if rv is None:
      abort(404)
    return rv

  def fast_count(self):
    """Fast counting, bypassing subqueries.

    By default SQLAlchemy count queries use subqueries (which are very slow
    on MySQL). This method is useful when counting over large numbers of rows
    (10k and more), as the following benchmark shows (~250k rows):

    .. code:: python

      In [1]: %time Cat.q.count()
      CPU times: user 0.01 s, sys: 0.00 s, total: 0.01 s
      Wall time: 1.36 s
      Out[1]: 281992L

      In [2]: %time Cat.c.scalar()
      CPU times: user 0.00 s, sys: 0.00 s, total: 0.00 s
      Wall time: 0.06 s
      Out[2]: 281992L

    """
    models = query_to_models(self)
    if len(models) != 1:
      # initial query is over more than one model
      # not clear how to implement the count in that case
      raise ValueError('Fast count unavailable for this query.')
    count_query = Query(func.count(), session=self.session)
    count_query = count_query.select_from(models[0])
    count_query._criterion = self._criterion
    return count_query.scalar()

  def random(self, n=1, dialect=None):
    """Returns n random model instances.

    :param n: the number of instances to return
    :type n: int
    :param dialect: the engine dialect (the implementation of random differs
      between MySQL and SQLite among others). By default will look up on the
      query for the dialect used. If no random function is available for the 
      chosen dialect, the fallback implementation uses total row count to 
      generate random offsets.
    :type dialect: str
    :rtype: model instances
    
    """
    if dialect is None:
      conn = self._connection_from_session()
      dialect = conn.dialect.name
    if dialect == 'mysql':
      rv = self.order_by(func.rand()).limit(n).all()
    elif dialect in ['sqlite', 'postgresql']:
      rv = self.order_by(func.random()).limit(n).all()
    else: # fallback implementation
      count = self.count()
      rv = [self.offset(randint(0, count - 1)).first() for _ in range(n)]
    if len(rv) == 1:
      return rv[0]
    return rv

  def to_dataframe(self, lazy=True, **kwargs):
    """Loads a dataframe with the records from the query and returns it.

    :param lazy: whether or not to load the underlying objects. If set to
      ``False``, the dataframe will be populated with the contents of
      ``to_json`` of the models, otherwise it will only contain the columns
      existing in the database (default behavior). If lazy is ``True``, this
      method also accepts the same keyword arguments as
      :func:`flasker.util.query_to_dataframe`. For convenience, if no
      ``exclude`` kwarg is specified, it will default to ``['_cache']``.
    :type lazy: bool
    :rtype: pandas.DataFrame

    Requires the ``pandas`` library to be installed.

    """
    if lazy:
      kwargs.setdefault('exclude', ['_cache'])
      return query_to_dataframe(
        self,
        connection=self._connection_from_session(),
        **kwargs
      )
    else:
      return DataFrame([model.to_json() for model in self])

  def to_records(self, **kwargs):
    """Raw execute of the query into a generator.

    :rtype: generator

    This method accepts the same keyword arguments as 
    :func:`flasker.util.query_to_records`.
    
    """
    return query_to_records(
      self,
      connection=self._connection_from_session(),
      **kwargs
    )


# Mutables

class _JSONEncodedType(TypeDecorator):

  """Base class for storing python mutables as a JSON.

  For mutability tracking, associate with a Mutable.

  The underlying immutable object is of kind ``sqlalchemy.types.unicodetext``.
  Note that it has a character limit so care is needed when storing very large
  objects.


  """

  impl = UnicodeText

  def process_bind_param(self, value, dialect):
    return dumps(value) if value else None

  def process_result_value(self, value, dialect):
    raise NotImplementedError()


class JSONEncodedDict(_JSONEncodedType):

  """Implements dictionary column field type for SQLAlchemy.

  This can be used as a Column type during table creation::

    some_column_name = Column(JSONEncodedDict)

  It also implements limited  mutability tracking to know when to update the
  database: set, del and update actions are tracked. If another method to
  update the dictionary is used, it will not automatically flag the
  dictionary for update (for example if a deeply nested key is updated).
  In such a case, the ``changed`` method needs the be called manually
  after the operation.

  """

  def process_result_value(self, value, dialect):
    return loads(value) if value else {}


class _MutableDict(Mutable, dict):

  """Used with JSONEncoded dict to be able to track updates.

  This enables the database to know when it should update the stored string
  representation of the dictionary. This is much more efficient than naive
  automatic updating after each query.

  """

  @classmethod
  def coerce(cls, key, value):
    """Convert plain dictionaries to Features."""
    if not isinstance(value, cls):
      if isinstance(value, dict):
        return cls(value)
      return Mutable.coerce(key, value) # this will raise an error
    else:
      return value

  def update(self, *args, **kwargs):
    """Detect dictionary update events and emit change events."""
    dict.update(self, *args, **kwargs)
    self.changed()
    
  def __setitem__(self, key, value):
    """Detect dictionary set events and emit change events."""
    dict.__setitem__(self, key, value)
    self.changed()
    
  def __delitem__(self, key):
    """Detect dictionary del events and emit change events."""
    dict.__delitem__(self, key)
    self.changed()
    
_MutableDict.associate_with(JSONEncodedDict)


class JSONEncodedList(_JSONEncodedType):

  """Implements list column field type for SQLAlchemy.

  This can be used as a Column type during table creation::

    some_column_name = Column(JSONEncodedList)

  Currently only set, delete, append and extend events are tracked. Others
  will require a call to ``changed`` to be persisted.

  """

  def process_result_value(self, value, dialect):
    return loads(value) if value else []


class _MutableList(Mutable, list):

  """Used with JSONEncoded list to be able to track updates.

  This enables the database to know when it should update the stored string
  representation of the dictionary. This is much more efficient than naive
  automatic updating after each query.

  """

  @classmethod
  def coerce(cls, key, value):
    """Convert plain dictionaries to Features."""
    if not isinstance(value, cls):
      if isinstance(value, list):
        return cls(value)
      return Mutable.coerce(key, value) # this will raise an error
    else:
      return value

  def append(self, *args, **kwargs):
    """Detect update events and emit change events."""
    list.append(self, *args, **kwargs)
    self.changed()

  def extend(self, *args, **kwargs):
    """Detect update events and emit change events."""
    list.extend(self, *args, **kwargs)
    self.changed()
    
  def __setitem__(self, index, value):
    """Detect set events and emit change events."""
    list.__setitem__(self, index, value)
    self.changed()
    
  def __delitem__(self, index):
    """Detect del events and emit change events."""
    list.__delitem__(self, index)
    self.changed()

_MutableList.associate_with(JSONEncodedList)




#!/usr/bin/env python

"""API Extension (requires the ORM extension).

This extension provides a base class to create API views.

Setup is as follows:

.. code:: python

  from flasker import current_project as pj
  from flasker.ext import API

  api = API(pj)

  View = api.View   # the base API view

Views can then be created for models as follows:

.. code:: python

  # Cat is a subclass of flasker.ext.orm.Base

  class CatView(View):

    __model__ = Cat

This view will create the following hooks:

* ``/cats``
* ``/cats/<id>``

Another slighly more complex example:

.. code:: python

  # House is a subclass of flasker.ext.orm.Base

  class HouseView(View):

    __model__ = House

    methods = ['GET', 'POST']
    subviews = ['cats']

This view will create the following hooks:

* ``/houses``
* ``/houses/<id>``
* ``/houses/<id>/cats``
* ``/houses/<id>/cats/<position>``

These are only two simple ways to add a view. Please refer to the documentation
for :class:`flasker.ext.api.BaseView` for the list of all available options.

"""

from flask import Blueprint, jsonify, request
from os.path import abspath, dirname, join
from sqlalchemy.orm import class_mapper
from time import time
from werkzeug.exceptions import HTTPException

from ..util.flask import make_view, View as _View, _ViewMeta
from ..util.helpers import uncamelcase
from ..util.sqlalchemy import Model, Query, query_to_models


class APIError(HTTPException):

  """Thrown when an API call is invalid.

  The following error codes can occur:

  * ``400 Bad Request`` if the request is badly formulated (wrong query
    parameters, invalid form data, etc.)
  * ``403 Forbidden`` if the request is not authorized by the server
  * ``404 Not Found`` if the request refers to a non-existent resource
  
  """

  def __init__(self, code, content):
    self.code = code
    self.content = content
    super(APIError, self).__init__(content)

  def __repr__(self):
    return '<APIError %r: %r>' % (self.message, self.content)


class API(object):

  """The main API object.

  :param project: the project against which the extension will be registered
  :type project: flasker.project.Project
  :param url_prefix: the blueprint URL prefix
  :type url_prefix: str
  :param index_view: whether or not to create a splash page for the api
  :type index_view: bool
  :param parser_options: dictionary of options to create the default request
    :class:`flasker.ext.api.Parser`
  :type parser_options: dict

  """

  def __init__(self, project, url_prefix='api', index_view=True,
               parser_options=None):

    parser_options = parser_options or {}

    blueprint = Blueprint(
      url_prefix,
      '%s.%s' % (project.flask.name, url_prefix),
      url_prefix='/%s' % url_prefix,
    )

    self.View = make_view(
      blueprint,
      view_class=View,
      parser=Parser(**parser_options)
    )

    @project.run_after_module_imports
    def api_after_imports(project):

      if index_view:
        @blueprint.route('/')
        def index():
          return jsonify({
            'available_endpoints': sorted(
              '%s (%s)' % (r.rule, ', '.join(str(meth) for meth in r.methods))
              for r in project.flask.url_map.iter_rules()
              if r.endpoint.startswith('%s.' % url_prefix)
            )
          })

      project.flask.register_blueprint(blueprint)
      project.logger.debug('api blueprint registered')

    project.logger.debug('api extension initialized')


class _ApiViewMeta(_ViewMeta):

  """To register classes with the API on definition.

  Automatically creates the ``endpoint``, ``base_url`` and ``rules`` for the
  view from the ``__model__`` attribute.

  Each route is then registered on the bound application (the current API
  blueprint here).
  
  """

  def __new__(cls, name, bases, dct):

    model = dct.get('__model__', None)

    if model is not None:
      if not issubclass(model, Model):
        raise ValueError('Api views can only be used with Orm models.')

      dct.setdefault('endpoint', model.__tablename__)
      base_url = dct.setdefault('base_url', model.__tablename__)

      collection_route = '/%s' % (base_url, )
      model_route = '/%s%s' % (
        base_url,
        ''.join('/<%s>' % k.name for k in class_mapper(model).primary_key)
      )

      dct['rules'] = {
        collection_route: ['GET', 'POST'],
        model_route: ['GET', 'PUT', 'DELETE'],
      }

    return super(_ApiViewMeta, cls).__new__(cls, name, bases, dct)


class View(_View):

  """Base API view.

  To customize, override the ``get``, ``post``, etc. methods.

  """

  __metaclass__ = _ApiViewMeta

  #: orm.Model class
  __model__ = None

  #: Base URL (will default to the model's tablename).
  base_url = None

  #: Allowed methods.
  methods = frozenset(['GET'])

  #: Request parser.
  parser = None

  #: Which relationship endpoints to create (these allow GET requests).
  #: Can be ``True`` (all relationships) or a list of relationship names.
  #: Only relationships with ``lazy`` set to ``'dynamic'``, ``'select'`` or
  #: ``True`` can have subroutes. All eagerly loaded relationships are simply
  #: available directly on the model.
  subviews = []

  @classmethod
  def register_view(cls, blueprint):
    """Create the URL routes for the view.
    
    Standard :class:`flasker.util.View` implementation plus subview support.
    
    """

    super(View, cls).register_view(blueprint)

    if cls.subviews:
      model = cls.__model__
      all_keys = set(
        model._get_relationships(
          lazy=['dynamic', True, 'select'],
          uselist=True
        ).keys() +
        model._get_association_proxies().keys()
      )

      if cls.subviews == True:
        keys = all_keys
      else:
        keys = set(cls.subviews)
        if keys - all_keys:
          raise ValueError('%s invalid for subviews' % (keys - all_keys, ))
        keys = all_keys & keys

      for key in keys:
        collection_route = '/%s%s/%s' % (
          cls.base_url,
          ''.join(
            '/<%s>' % k.name for k in class_mapper(model).primary_key
          ),
          key,
        )
        model_route = '/%s%s/%s/<position>' % (
          cls.base_url,
          ''.join(
            '/<%s>' % k.name for k in class_mapper(model).primary_key
          ),
          key
        )
        make_view(
          blueprint,
          view_class=_RelationshipView,
          view_name='%s_%s' % (cls.endpoint, key),
          __model__=model,
          __assoc_key__=key,
          parser=cls.parser,
          endpoint='%s_%s' % (cls.endpoint, key),
          methods=['GET', ],
          rules={
            collection_route: ['GET', ],
            model_route: ['GET', ],
          },
        )

  def get(self, **kwargs):
    """GET request handler."""
    if kwargs:
      model = self.__model__.retrieve(from_key=True, **kwargs)
      if not model:
        raise APIError(404, 'Not found')
      return self.parser.jsonify(model)
    else:
      return self.parser.jsonify(self.__model__.q)

  def post(self):
    """POST request handler."""
    if not self.validate(json):
      raise APIError(400, 'Invalid POST parameters')
    model = self.__model__(**request.json)
    model.flush()
    return self.parser.jsonify(model)

  def put(self, **kwargs):
    """PUT request handler."""
    model = self.__model__.retrieve(from_key=True, **kwargs)
    if not model:
      raise APIError(404, 'Not found')
    if not self.validate(json, model):
      raise APIError(400, 'Invalid PUT parameters')
    for k, v in request.json.items():
      setattr(model, k, v)
    return self.parser.jsonify(model)

  def delete(self, **kwargs):
    """DELETE request handler."""
    model = self.__model__.retrieve(from_key=True, **kwargs)
    if not model:
      raise APIError(404, 'Not found')
    pj.session.delete(model)
    return self.parser.jsonify(model)

  def validate(self, json, model=None):
    """Validation method.

    :param json: a dictionary of attributes
    :type json: dict
    :param model: ``None`` if it is POST request, and the model instance to be
      modified if it is a PUT request.
    :type model: None or flasker.ext.orm.BaseModel
    :rtype: bool

    This method is called on each POST and PUT request. Override it to
    implement your own validation logic: return ``True`` when the input is
    valid and ``False`` otherwise. Default implementation accepts everything.

    """
    return True
  

class _RelationshipView(_View):

  """Relationship View."""

  __model__ = None
  __assoc_key__ = None

  def get(self, **kwargs):
    """GET request handler."""
    position = kwargs.pop('position', None)
    parent = self.__model__.retrieve(from_key=True, **kwargs)
    if not parent:
      raise APIError(404, 'Parent not found')
    collection =  getattr(parent, self.__assoc_key__)

    if position:
      position = int(position) - 1 # model_position is 1 indexed
      if isinstance(collection, Query):
        model = collection.offset(position).limit(1).first()
      else:
        collection = collection[position:(position + 1)]
        model = collection[0] if collection else None
        if not model:
          raise APIError(404, 'Not found')
      return self.parser.jsonify(model)

    else:
      return self.parser.jsonify(collection)


class Parser(object):

  """The request parameter parser.

  :param default_depth: the default depth models are jsonified to. ``0`` yields 
    an empty dictionary
  :type default_depth: int
  :param max_depth: the maximum depth allowed in a query. ``0`` means no limit.
  :type max_depth: int
  :param default_limit: the default number of results returned per query 
  :type default_limit: int
  :param max_limit: the maximum number of results returned by a query. ``0`` 
    means no limit.
  :type max_limit: int
  :param sep: the separator used for filters and sort parameters
  :type sep: str

  This class has a single method :meth:``jsonify`` which is used to parse a
  model or collection and return the serialized response.
  
  """

  def __init__(self, default_depth=1, max_depth=0, default_limit=20,
               max_limit=0, sep=';'):
    self.options = {
      'default_depth': default_depth,
      'max_depth': max_depth,
      'default_limit': default_limit,
      'max_limit': max_limit,
      'sep': sep,
    }

  def jsonify(self, data, data_key='data', meta_key='meta',
    include_request=True, include_time=True, include_matches=True, **kwargs):
    """Parses the data and returns the serialized response.

    :param data: data. At this time, only instances, and lists of instances of
      ``flasker.util.sqlalchemy.Model``, along with instances of 
      ``flasker.util.sqlalchemy.Query`` are valid.
    :type data: model or collection
    :param data_key: key where the serialized data will go
    :type data_key: str
    :param meta_key: key where the metadata will go
    :type meta_key: str
    :param include_request: whether or not to include the issued request
      information
    :type include_request: bool
    :param include_time: whether or not to include processing time
    :type include_time: bool
    :param include_matches: whether or not to include the total number of
      results from the data (useful if ``data`` is a collection)
    :type include_matches: bool
    :rtype: Flask response
    
    Any keyword arguments will be included with the metadata.
    
    """
    depth = request.args.get('depth', self.options['default_depth'], int)
    max_depth = self.options['max_depth']
    if max_depth:
      depth = min(depth, max_depth)

    start = time()

    if isinstance(data, Model):
      data = data.to_json(depth=depth)
      match = 1
    else:
      col, matches = self._get_collection(data)
      data = [e.to_json(depth=depth) for e in col if e]
      match = {'total': matches, 'returned': len(data)}

    rv = {data_key: data, meta_key: kwargs}

    if include_matches:
      rv[meta_key]['matches'] = match
    if include_request:
      rv[meta_key]['request'] = {
        'base_url': request.base_url,
        'method': request.method,
        'values': request.values,
      }
    if include_time:
      rv[meta_key]['parsing_time'] = time() - start

    return jsonify(rv)

  def _get_collection(self, collection):
    """Parse query and return JSON.

    :param collection: the query or list to be transformed to JSON
    :type collection: flasker.ext.orm.Query, list
    :rtype: tuple

    Returns a tuple ``(collection, match)``:

    * ``collection`` is the filtered, sorted, offsetted, limited collection.
    * ``match`` is the total number of results from the filtered query 

    """
    model = self._get_model_class(collection)
    raw_filters = request.args.getlist('filter')
    raw_sorts = request.args.getlist('sort')
    offset = request.args.get('offset', 0, int)
    limit = request.args.get('limit', self.options['default_limit'], int)
    max_limit = self.options['max_limit']
    if max_limit:
      limit = min(limit, max_limit) if limit else max_limit

    if isinstance(collection, Query):

      sep = self.options['sep']

      for raw_filter in raw_filters:
        try:
          key, op, value = raw_filter.split(sep, 3)
        except ValueError:
          raise APIError(400, 'Invalid filter: %s' % raw_filter)
        column = getattr(model, key, None)
        if not column: # TODO check if is actual column
          raise APIError(400, 'Invalid filter column: %s' % key)
        if op == 'in':
          filt = column.in_(value.split(','))
        else:
          try:
            attr = filter(
              lambda e: hasattr(column, e % op),
              ['%s', '%s_', '__%s__']
            )[0] % op
          except IndexError:
            raise APIError(400, 'Invalid filter operator: %s' % op)
          if value == 'null':
            value = None
          filt = getattr(column, attr)(value)
        collection = collection.filter(filt)

      for raw_sort in raw_sorts:
        try:
          key, order = raw_sort.split(sep)
        except ValueError:
          raise APIError(400, 'Invalid sort: %s' % raw_sort)
        if not order in ['asc', 'desc']:
          raise APIError(400, 'Invalid sort order: %s' % order)
        column = getattr(model, key, None)
        if column:
          collection = collection.order_by(getattr(column, order)())
        else:
          raise APIError(400, 'Invalid sort column: %s' % key)

      matches = collection.fast_count()
      if offset:
        collection = collection.offset(offset)
      if limit:
        collection = collection.limit(limit)

    else:
      if raw_filters or raw_sorts:
        raise APIError(400, 'Filter and sorts not implemented for lists')

      matches = len(collection)
      if limit:
        collection = collection[offset:(offset + limit)]
      else:
        collection = collection[offset:]

    return collection, matches

  def _get_model_class(self, collection):
    """Return corresponding model class from collection."""
  
    if isinstance(collection, Query):
      models = query_to_models(collection)

      # only tested for _BaseQueries and associated count queries
      assert len(models) < 2, 'Invalid query'

      if not len(models):
        # this is a count query
        return collection._select_from_entity
      else:
        # this is a Query
        return models[0]

    else:
      return collection[0].__class__




#!/usr/bin/env python

from celery import Celery
from flask import Flask
from itertools import repeat
from json import loads
from nose.tools import ok_, eq_, nottest, raises, timed
from os import chdir, pardir
from os.path import abspath, dirname, exists, join
from requests import ConnectionError, get
from sqlalchemy.orm.scoping import scoped_session
from subprocess import Popen, PIPE
from threading import Thread
from time import sleep, time

from flasker.project import *


class Test_Project(object):

  def setup(self):
    self.cp = abspath(
      join(
        dirname(__file__),
        pardir,
        pardir,
        'examples',
        'basic',
        'default.cfg'
      )
    )
    if not exists(self.cp):
      raise Exception('Missing configuration file')

  def teardown(self):
    Project._Project__state = {}

  @staticmethod
  def check_thread(creator, opj):
    pj = creator()
    eq_(pj.__dict__, opj.__dict__)
    eq_(pj.flask, opj.flask)
    eq_(pj.celery, opj.celery)
    eq_(pj.session, opj.session)

  def test_config_path(self):
    pj = Project(self.cp)
    eq_(self.cp, pj.conf_path)

  @raises(ProjectImportError)
  def test_unique_project(self):
    pj = Project(self.cp)
    another = Project('some/other/path.cfg')

  @raises(ProjectImportError)
  def test_empty_config_path(self):
    pj = Project()

  def test_components(self):
    pj = Project(self.cp)
    eq_(pj._flask, None)
    eq_(pj._celery, None)
    eq_(pj._session, None)
    eq_(type(pj.flask), Flask)
    eq_(type(pj.celery), Celery)
    eq_(type(pj.session), scoped_session)

  def test_borg_pattern(self):
    pj = Project(self.cp)
    another = Project()
    eq_(another.__dict__, pj.__dict__)

  def test_proxy(self):
    pj = Project(self.cp)
    eq_(current_project.__dict__, pj.__dict__)

  def test_threaded_project(self):
    pj = Project(self.cp)
    th = Thread(target=self.check_thread, args=(lambda: Project(), pj))
    th.start()
    th.join()

  def test_threaded_proxy(self):
    pj = Project(self.cp)
    th = Thread(target=self.check_thread, args=(lambda: current_project, pj))
    th.start()
    th.join()

  def test_app_server(self):
    pj = Project(self.cp)
    client = pj.flask.test_client()
    eq_(client.get('/').data, 'Hello World!')


# class Test_ConsoleTool(object):
# 
#   def setup(self):
#     chdir(abspath(
#       join(
#         dirname(__file__),
#         pardir,
#         pardir,
#         'examples',
#         'basic'
#       )
#     ))
#     self.sps = []
# 
#   def teardown(self):
#     for sp in self.sps:
#       if sp.poll() is not None:
#         sp.terminate()
# 
#   def open_subprocess(self, command):
#     sp = Popen(command)
#     # self.sps.append(sp)
#     return sp
# 
#   @timed(3)
#   def test_server(self):
#     sp = self.open_subprocess(['flasker', 'server', '-p', '5050'])
#     for _ in repeat(None, 4):
#       try:
#         sleep(1)
#         eq_(get('http://localhost:5050').json(), {'message': 'Welcome!'})
#         break
#       except ConnectionError:
#         pass
#     sp.terminate()
    



#!/usr/bin/env python

"""Thin wrapper around some sklearn algorithms.

Allows for automatic parameter saving and recovery. Saving fitted algorithms
for later reuse (convenient for parallel computing).

Test results are persisted along with a few metrics.

Currently supported:

* ``sklearn.linear_model.logistic.LogisticRegression``

Example usage:

.. code:: python

  >>> clf = Classifier.from_engine(LogisticRegression, C=3)
  >>> fitted_clf = clf.train(X_train, y_train)
  >>> test = fitted_clf.test(X_test, y_test)
  >>> test.precision
  0.9576324

"""

from datetime import datetime
from time import time
from json import loads, dumps
from os import makedirs
from os.path import exists, join
from sklearn.externals import joblib
from sqlalchemy import (Boolean, Column, Float, ForeignKey, Integer, String,
  DateTime, Numeric, Text)
from sys import modules


class SKL(object):

  def __init__(self, project, orm, table_prefix='skl_', folderpath=None,
               param_class=None, engine_class=None):

    self.table_prefix = table_prefix
    self.folderpath = folderpath or table_prefix.rstrip('_')
    self.root_dir = join(project.root_dir, self.folderpath)
    if not exists(self.root_dir): makedirs(self.root_dir)

    Param, Engine = _create_base_models(orm)

    self.Param = type(
      '%sParam' % self.folderpath.capitalize(),
      (Param, param_class or object),
      {
        '__tablename__': '%sparams' % self.table_prefix,
      }
    )

    self.Engine = type(
      '%sEngine' % self.folderpath.capitalize(),
      (Engine, engine_class or object),
      {
        '__tablename__': '%sengines' % self.table_prefix,
        'folderpath': self.folderpath,
        'param_id': Column(ForeignKey('%s.id' % self.Param.__tablename__)),
        'param': orm.relationship(
          self.Param,
          backref=orm.backref('engines', lazy='dynamic')
        )
      }
    )

  def save(self, engine, description, dump_engine=True):
    """Wrap an sklearn algorithm.

    :param engine: sklearn algorithm or algorithm factory
    :type engine: varies
    :rtype: flasker.ext.skl.UnfittedClassifier

    """
    param, flag = self.Param.retrieve(
      type='%s.%s' % (
        engine.__module__,
        engine.__class__.__name__
      ),
      **self.Param._serialize_params(**engine.get_params())
    )
    _engine = self.Engine(
      description=description,
      param=param
    )
    _engine.flush()
    if dump_engine:
      _engine.save(engine)
    return _engine

  def load(self, engine_id=None, param_id=None):
    """Load an unfitted classifier from its parameter id.

    :param param_id: the id of the parameter to use to load the engine
    :type param_id: int
    :rtype: flasker.ext.skl.UnfittedClassifier

    The parameter is persisted in the database and its kwargs are passed to 
    the classifier's constructor.

    """
    if engine_id:
      return self.Engine.q.get(engine_id).load()
    if param_id:
      return self.Param.q.get(param_id).create_engine()


def _create_base_models(orm):

  Model = orm.Model
  backref = orm.backref
  relationship = orm.relationship

  class Param(Model):

    __abstract__ = True

    _cache = None
    id = Column(Integer, primary_key=True)
    type = Column(String(64))
    description = Column(Text)
    added = Column(DateTime, default=datetime.now)
    penalty = Column(String(32))
    dual = Column(Boolean)
    C = Column(Numeric(precision=64, scale=30))
    fit_intercept = Column(Boolean)
    intercept_scaling = Column(Numeric(precision=64, scale=30))
    random_state = Column(Boolean)
    class_weight = Column(Text)
    tol = Column(Numeric(precision=64, scale=30))
    n_estimators = Column(Integer)
    criterion = Column(String(32))
    max_features = Column(String(32))
    max_depth = Column(Integer)
    min_samples_split = Column(Integer)
    min_samples_leaf = Column(Integer)
    min_density = Column(Numeric(precision=64, scale=30))
    bootstrap = Column(Boolean)
    oob_score = Column(Boolean)
    n_jobs = Column(Integer)
    verbose = Column(Boolean)
    compute_importances = Column(Boolean)

    @classmethod
    def _serialize_params(cls, **kwargs):
      if 'class_weight' in kwargs:
        class_weight = kwargs['class_weight']
        if isinstance(class_weight, dict):
          kwargs['class_weight'] = dumps(class_weight)
      if 'max_features' in kwargs:
        kwargs['max_features'] = str(kwargs['max_features'])
      return kwargs

    @classmethod
    def _unserialize_params(cls, **kwargs):
      if 'class_weight' in kwargs and kwargs['class_weight']:
        try:
          kwargs['class_weight'] = {
            int(k): v for k, v in loads(kwargs['class_weight']).items()
          }
        except ValueError:
          pass # it is a simple string (probably 'auto')
      if 'max_features' in kwargs and kwargs['max_features']:
        try:
          kwargs['max_features'] = int(kwargs['max_features'])
        except ValueError:
          pass
      return kwargs

    def create_engine(self):
      """Create new engine."""
      module, cls = self.type.rsplit('.', 1)
      __import__(module)
      engine_factory = getattr(modules[module], cls)
      valid_kwargs = engine_factory._get_param_names()
      kwargs = self._unserialize_params(**self.to_json())
      return engine_factory(
        **{k: v for k, v in kwargs.items() if k in valid_kwargs}
      )


  class Engine(Model):

    __abstract__ = True

    _cache = None
    id = Column(Integer, primary_key=True)
    description = Column(Text)
    added = Column(DateTime, default=datetime.now)
    folderpath = None

    def load(self):
      joblib.load(join(self.folderpath, '%s.joblib.pkl' % self.id))

    def save(self, engine):
      joblib.dump(engine, join(self.folderpath, '%s.joblib.pkl' % self.id))


  return Param, Engine



#!/usr/bin/env python

"""Nested dictionary helpers."""

from collections import Mapping
from copy import deepcopy


def depth(dct):
  """Depth of a dictionary.

  :param dct: dictionary
  :type dct: dict
  :rtype: int
  
  """
  values = [
      value
      for value in dct.itervalues()
      if isinstance(value, dict)
  ]
  return max(depth(value) for value in values) + 1 if values else 1

def width(dct):
  """Width of a dictionary.

  :param dct: dictionary
  :type dct: dict
  :rtype: int
  
  """
  values = [
      value
      for value in dct.itervalues()
      if isinstance(value, dict)
  ]
  return sum(width(value) for value in values) + len(dct) - len(values)

def flatten(dct, sep='_', prefix=''):
  """Flatten a dictionary.

  :param dct: dictionary
  :type dct: dict
  :param sep: the separator used when concatenating keys
  :type sep: str
  :param prefix: a prefix to add to all new keys
  :type prefix: str
  :rtype: dict:

  .. note::

    All keys in the dictionary must be strings.

  """
  items = []
  for key, value in dct.iteritems():
    k = prefix + sep + key if prefix else key
    if isinstance(value, dict) and value:
      items.extend(flatten(value, sep, k).items())
    else:
      items.append((k, value))
  return dict(items)

def unflatten(dct, sep='_', cname='all'):
  """Unflatten a dictionary.

  :param dct: dictionary
  :type dct: dict
  :param sep: the separator used to split keys
  :type sep: str
  :param cname: a key name used when a value would be added on an already
    expanded dictionary. A simple example is when trying to unflatten 
    ``{'a': 1, 'a_b': 2}``: there is ambiguity on where to store the value
    for key ``'a'`` because it already contains the dictionary ``{'b': 2}``.
    This is resolved creating a new key ``cname`` in this latter dictionary.
  :type cname: str
  :rtype: dict:

  """
  result = {}
  keys = []
  for key in dct.iterkeys():
    keys.append(key.split(sep))
  keys.sort(key=len, reverse=True)
  for key in keys:
    d = result
    for part in key[:-1]:
      if part not in d:
        d[part] = {}
      d = d[part]
    if key[-1] in d:
      d[key[-1]][cname] = dct[sep.join(key)]
    else:
      d[key[-1]] = dct[sep.join(key)]
  return result

def update(a, b, copy=False):
  """Update for nested dictionaries.

  :param copy: whether or not to do a deepcopy of the dictionary before
    updating.
  :type copy: bool
  :rtype: dict

  """
  if copy:
    a = deepcopy(a)
  for k, v in b.iteritems():
    if isinstance(v, Mapping):
      a[k] = update(a.get(k, {}), v)
    else:
      a[k] = b[k]
  return a




  #!/usr/bin/env python

"""Project module.

This module defines:

* the :class:`flasker.project.Project` class which contains
  all the logic between the Flask and Celery applications and the SQLAlchemy
  sessions.

* the ``current_project`` proxy

For convenience, both these variables are also available directly in the
``flasker`` namespace.

.. note::

  In most cases :class:`flasker.project.Project` will not need to be
  instantiated explicitely (the console tool handles the setup) and will only
  be accessed via the ``current_project`` proxy. In some cases however the
  constructor can be called to create a project (for example from an IPython
  notebook or to use a separate WSGI server).

  .. code:: python

    from flasker import Project

    # instantiating the project
    pj = Project('path/to/config.cfg')

    # the application that would be passed to a WSGI server
    application = pj.flask

"""

from logging import getLogger, NullHandler, StreamHandler, DEBUG
from os.path import abspath, dirname, join, sep, split, splitext
from sys import path
from werkzeug.local import LocalProxy

from .util.helpers import parse_config


class ProjectImportError(Exception):

  pass


class Project(object):

  """Project class.

  :param conf_path: path to the configuration file. The following sections
    are special: ``PROJECT``, ``ENGINE``, ``FLASK``, ``CELERY``. All but 
    ``PROJECT`` are optional. See below for a list of available options in each
    section.
  :type conf_path: str

  The following options are available each section of the configuration file
  (as a convenience, parameters stored as JSON strings are also accepted,
  which can be useful for example to configure Celery queues):

  * ``PROJECT``

    * ``MODULES``: comma separated list of the project's modules. They must be
      importable from the configuration file's folder.
    * ``DEBUG``
    * ``DISABLE_FLASK``
    * ``DISABLE_CELERY``

  * ``ENGINE``

    * ``URL``: the url to the database
    * any valid arguments to ``sqlalchemy.create_engine``

  * ``SESSION``

    * ``SMARTCOMMIT``: if ``True`` (default), all database transactions
      will be committed after each Flask app request and Celery task
      completion. If ``False`` the session will simply be removed.
    * any valid arguments to ``sqlalchemy.orm.session_maker``

  * ``FLASK``

    * ``ROOT_FOLDER``: path to the Flask application's root folder
      relative to the configuration file (defaults to ``app``).
    * ``STATIC_FOLDER``: the application's ``static_folder`` relative to
      the application's root folder (defaults to ``static``).
    * ``TEMPLATE_FOLDER``: the application's ``template_folder`` relative
      to the application's root folder (defaults to ``templates``).
    * any valid Flask configuration option

  * ``CELERY``

    * ``MAIN`` 
    * any valid Celery configuration option
  
  """

  #: Default configuration
  default_conf = {
    'PROJECT': {
      'MODULES':          '',
      'DEBUG':            False,
      'DISABLE_FLASK':    False,
      'DISABLE_CELERY':   False,
    },
    'ENGINE': {
      'URL':              'sqlite://',
    },
    'SESSION': {
      'SMARTCOMMIT':      True,
    },
    'FLASK': {
      'NAME':             'app',
      'ROOT_FOLDER':      'app',
      'STATIC_FOLDER':    'static',
      'TEMPLATE_FOLDER':  'templates',
    },
    'CELERY': {
      'MAIN':             '__main__',
    },
  }

  #: Dictionary of configuration values
  conf = None

  #: Path to current configuration file
  conf_path = None

  #: Logger
  logger = None

  _flask = None
  _celery = None
  _session = None

  __state = {}

  def __init__(self, conf_path=None):

    self.__dict__ = self.__state

    if conf_path is None:

      if not self.conf_path:
        raise ProjectImportError('Project instantiation outside the Flasker '
                                 'command line tool requires a '
                                 'configuration file path.')

    else:

      if self.conf_path and abspath(conf_path) != self.conf_path:
        raise ProjectImportError('Cannot instantiante projects for different '
                                 'configuration files in the same process.')

      elif not self.conf_path:

        # load configuration
        self.conf = parse_config(
          conf_path,
          default=self.default_conf,
          case_sensitive=True
        )
        self.conf_path = abspath(conf_path)

        # load logger
        self.logger = getLogger(__name__)
        if self.conf['PROJECT']['DEBUG']:
          self.logger.setLevel(DEBUG)
          self.logger.addHandler(StreamHandler())
        else:
          self.logger.addHandler(NullHandler())

        # load all project modules
        self._funcs = []
        path.append(dirname(self.conf_path))
        project_modules = [
          module_name.strip()
          for module_name in self.conf['PROJECT']['MODULES'].split(',')
        ]
        for module_name in project_modules:
          __import__(module_name)
        self.logger.debug(
          '%s modules imported (%s)' % (
            len(project_modules),
            ', '.join(project_modules),
          )
        )
        for func in self._funcs:
          func(self)
        self.logger.debug(
          '%s handlers found and run (%s)' % (
            len(self._funcs), 
            ', '.join(func.__name__ for func in self._funcs),
          )
        )

  def __repr__(self):
    return '<Project %r>' % (self.conf_path, )

  @property
  def flask(self):
    """Flask application.

    Lazily initialized.

    """
    if self._flask is None and not self.conf['PROJECT']['DISABLE_FLASK']:

      from flask import Flask

      flask_app = Flask(
        self.conf['FLASK']['NAME'],
        static_folder=self.conf['FLASK']['STATIC_FOLDER'],
        template_folder=self.conf['FLASK']['TEMPLATE_FOLDER'],
        instance_path=join(
          dirname(self.conf_path),
          self.conf['FLASK']['ROOT_FOLDER'],
        ),
        instance_relative_config=True,
      )

      flask_app.config.update({
        k: v
        for k, v in self.conf['FLASK'].items()
        if not k in self.default_conf['FLASK']
      })

      self.logger.debug('flask app loaded')
      self._flask = flask_app
    return self._flask

  @property
  def celery(self):
    """Celery application.

    Lazily initialized.

    """
    if self._celery is None and not self.conf['PROJECT']['DISABLE_CELERY']:

      from celery import Celery
      from celery.task import periodic_task

      celery_app = Celery(self.conf['CELERY']['MAIN'])

      celery_app.conf.update({
        k: v
        for k, v in self.conf['CELERY'].items()
        if not k in self.default_conf['CELERY']
      })

      # proxy for easy access
      celery_app.periodic_task = periodic_task

      # maybe not required with lazy session initialization
      # TODO: check this
      # @worker_process_init.connect
      # def create_worker_connection(*args, **kwargs):
      #   self._create_session()

      self.logger.debug('celery app loaded')
      self._celery = celery_app
    return self._celery

  @property
  def session(self):
    """SQLAlchemy scoped sessionmaker.

    Lazily initialized.

    """
    if self._session is None:

      from celery.signals import task_postrun
      from flask.signals import request_tearing_down

      from sqlalchemy import create_engine  
      from sqlalchemy.exc import InvalidRequestError
      from sqlalchemy.orm import scoped_session, sessionmaker

      engine = create_engine(
        self.conf['ENGINE']['URL'],
        **{
          k.lower(): v 
          for k, v in self.conf['ENGINE'].items()
          if not k in self.default_conf['ENGINE']
        }
      )
      session = scoped_session(
        sessionmaker(
          bind=engine,
          **{
            k.lower(): v
            for k, v in self.conf['SESSION'].items()
            if not k in self.default_conf['SESSION']
          }
        )
      )

      task_postrun.connect(_remove_session)
      request_tearing_down.connect(_remove_session)

      self.logger.debug('session loaded')
      self._session = session
    return self._session

  def run_after_module_imports(self, func):
    """Hook to run a function right after all project modules are imported.

    :param func: the function to be called right before startup. It will be
      passed the project as single argument.
    :type func: callable

    This decorator can be used to run functions after all the components of
    the project have been created.
    
    """
    self._funcs.append(func)

  def _remove_session(self):
    """Remove database connections."""
    try:
      if self.conf['SESSION']['SMARTCOMMIT']:
        self.session.commit()
        self.logger.debug('session committed')
    except InvalidRequestError as e:
      self.session.rollback()
      self.session.expunge_all()
      raise e
    finally:
      self.session.remove()
      self.logger.debug('session removed')

  def _reset(self):
    """Reset current project."""
    self.__class__.__state = {}


#: Proxy to the current project
current_project = LocalProxy(Project)

def _remove_session(*args, **kwargs):
  """Globally namespaced function for signals to work."""
  current_project._remove_session()




#!/usr/bin/env python

from tempfile import NamedTemporaryFile
from nose.tools import eq_, raises

from flasker.util.helpers import *
from flasker.util.mixins import *
from flasker.util.ndict import *


def test_convert_auto():
  examples = [
    ('1', 1), ('None', None), ('hi', 'hi'), ('true', True), ('False', False),
    ('1.432', 1.432), ('1e-3', 1e-3)
  ]
  for example in examples:
    yield check_convert, example, None, False

def test_convert_json():
  examples = [
    ('1', 1), ('None', None), ('hi', 'hi'), ('true', True), ('False', False),
    ('1.432', 1.432), ('1e-3', 1e-3), ('null', None), ('{"a": 2}', {'a': 2}),
  ]
  for example in examples:
    yield check_convert, example, None, True

def test_convert_manual():
  all_examples = {
    'int': [
      ('1', 1),
    ],
    'float': [
      ('2.3', 2.3),
    ],
    'bool': [
      ('True', True),
      ('1', True),
      ('false', False),
      ('0', False),
    ],
    'unicode': [
      ('hello', u'hello'),
    ],
    'str': [
      ('string', 'string'),
    ],
    'json': [
      ('null', None),
      ('{"a": 23}', {'a': 23}),
    ]
  }
  for rtype, examples in all_examples.items():
    for example in examples:
      yield check_convert, example, rtype, False

def check_convert(example, rtype, allow_json):
  eq_(convert(example[0], rtype=rtype, allow_json=allow_json), example[1])

def test_partition_list():
  col = range(103)
  for i, (q, p) in enumerate(partition(col, size=5)):
    eq_(p.offset, 5 * i)
    eq_(p.limit, 5)
    if i < 20:
      eq_(len(q), 5)
    else:
      eq_(len(q), 3)

def test_prod():
  eq_(prod(range(1, 5)), 2 * 3 * 4)

def test_uncamelcase():
  eq_(uncamelcase('CalvinAndHobbes'), 'calvin_and_hobbes')

def test_parse_config():
  f = NamedTemporaryFile()
  f.write('[SECTION_A]\n'
          'FIRST = 3\n'
          'SECOND = hi')
  f.seek(0)
  eq_(parse_config(f), {'SECTION_A': {'first': 3, 'second': 'hi'}})
  f.seek(0)
  eq_(
    parse_config(f, case_sensitive=True),
    {'SECTION_A': {'FIRST': 3, 'SECOND': 'hi'}}
  )
  f.seek(0)
  eq_(
    parse_config(f, default={'SECTION_A': {'third': 0.0}, 'C': {'O': 2}}),
    {'SECTION_A': {'first': 3, 'second': 'hi', 'third': 0.0}, 'C': {'O': 2}}
  )
  f.close()


class Test_Dict(object):

  def setup(self):
    self.examples = [
      {
        'u': {
          'a': 1,
          'b': {
            'c': None,
            'd': 'Cat',
          },
          'c': {
            '1': {
              'A': []
            }
          },
        },
        'f': {
          'a': 1,
          'b_c': None,
          'b_d': 'Cat',
          'c_1_A': [],
        }
      },
      {
        'u': {
          'd': {
            'all': 0,
            'e': -1,
          },
        },
        'f': {
          'd': 0,
          'd_e': -1,
        }
      }
    ]

  def test_depth(self):
    for example, d in zip(self.examples, [3,2]):
      eq_(depth(example['u']), d)
      eq_(depth(example['f']), 1)

  def test_width(self):
    for example, w in zip(self.examples, [4,2]):
      for v in example.values():
        eq_(width(v), w)
  
  def test_flatten(self):
    example = self.examples[0]
    eq_(flatten(example['u']), example['f'])

  def test_unflatten(self):
    for example in self.examples:
      eq_(unflatten(example['f']), example['u'])

  def test_update(self):
    a = {'a': 1, 'b': {'c': 0}}
    b = {'a': 2, 'b': {'d': 1}}
    c = update(a, b)
    eq_(c, {'a': 2, 'b': {'c': 0, 'd': 1}})
    eq_(a, c)
    eq_(b, {'a': 2, 'b': {'d': 1}})

  def test_update_with_copy(self):
    a = {'a': 1, 'b': {'c': 0}}
    b = {'a': 2, 'b': {'d': 1}}
    c = update(a, b, copy=True)
    eq_(a, {'a': 1, 'b': {'c': 0}})
    eq_(b, {'a': 2, 'b': {'d': 1}})
  

class Test_Cacheable(object):

  def setup(self):

    class Example(Cacheable):

      @Cacheable.cached_property
      def number(self):
        return 10

      @Cacheable.cached_property
      def another(self):
        return 48

    self.ex = Example()

  def test_set_cache(self):
    eq_(self.ex.number, 10)
    self.ex.number = 2
    eq_(self.ex.number, 2)
    self.ex.number = 10

  def test_refresh_cache(self):
    self.ex.number = 3
    self.ex.refresh_cache(['another'])
    eq_(self.ex.number, 3)
    self.ex.refresh_cache()
    eq_(self.ex.number, 10)

  def test_refresh_cache_expiration(self):
    self.ex.number = 3
    self.ex.refresh_cache(expiration=100)
    eq_(self.ex.number, 3)
    self.ex.refresh_cache()
    eq_(self.ex.number, 10)

  @raises(AttributeError)
  def test_refresh_cache_error(self):
    self.ex.refresh_cache(['number2'])

  def test_view_cache(self):
    self.ex.refresh_cache()
    eq_(set(self.ex.view_cache().keys()), set(['number', 'another']))


def test_to_json():

  class Foo(Jsonifiable):
    def __init__(self, n, nested=0):
      self.n = n
      if nested > 0:
        self.d = {str(n): Foo(n + 1, nested - 1)}
        self.f = Foo(n + 1, nested - 1)
        self.l = [Foo(n + 1, nested - 1)]
        self._p = 33
  foo = Foo(0, 2)
  j = foo.to_json()
  eq_(j, {'n': 0, 'd': {'0': {}}, 'f': {}, 'l': [{}]})
  j['n'] = 1
  j['d'] = {'1': {}}
  eq_(foo.to_json(depth=2), {'n': 0, 'd': {'0': j}, 'f': j, 'l': [j]})
  Foo.__json__ = ['n']
  eq_(foo.to_json(depth=2), {'n': 0})




#!/usr/bin/env python

from logging import getLogger
from time import time

from .helpers import to_json


class Jsonifiable(object):

  """JSONification mixin."""

  @property
  def __json__(self):
    """Default implementation of the attributes to jsonify.

    This is relatively slow (because it is evaluated for each jsonify call).
    Consider overwriting it for better performance.
    
    """
    return [
      varname for varname in dir(self)
      if not varname.startswith('_')
      if not callable(getattr(self, varname))
    ]

  def to_json(self, depth=1):
    """Returns all keys and properties of an instance in a dictionary.

    :param depth:
    :type depth: int
    :rtype: dict

    """
    rv = {}
    if depth < 1:
      return rv
    for varname in self.__json__:
      try:
        rv[varname] = to_json(getattr(self, varname), depth)
      except ValueError as e:
        rv[varname] = e.message
    return rv


class Loggable(object):

  """Convenient logging mixin.

  This implements the main logging methods directly on the class instance. For
  example, this allows something like::

    instance.info('Some message.')

  The instance's ``__str__`` is prepended to the message for easier debugging.

  Note that this class doesn't override `` __getattr__`` to preserve exception
  context. Otherwise the line where the inexistent attribute was accessed will
  be lost.

  """

  def _logger(self, message, loglevel):
    if not hasattr(self, '__logger__'):
      self.__logger__ = getLogger(self.__module__)
    action = getattr(self.__logger__, loglevel)
    return action('%s :: %s' % (self, message))

  def debug(self, message):
    """Debug level message."""
    return self._logger(message, 'debug')

  def info(self, message):
    """Info level message."""
    return self._logger(message, 'info')

  def warn(self, message):
    """Warn level message."""
    return self._logger(message, 'warn')

  def error(self, message):
    """Error level message."""
    return self._logger(message, 'error')


class Cacheable(object):

  """Mixin to support cacheable properties.
  
  Implements a few cache maintenance utilities.
  
  """

  _cache = None

  def _get_cached_properties(self):
    return [
        varname
        for varname in dir(self.__class__)
        if isinstance(getattr(self.__class__, varname), _CachedProperty)
    ]

  def refresh_cache(self, names=None, expiration=0, remove_deleted=True):
    """Refresh cached properties.

    :param names: list of cached property names to refresh. If specified, only
      these will be refreshed.
    :type names: iterable
    :param expiration: if specified, only properties of age greater than this
      value will be refreshed.
    :type expiration: int
    :param remove_deleted: if ``True``, cached properties that aren't defined
      but are still present in the cache will be removed (useful especially for
      persistent caches)
    :type remove_deleted: bool

    """
    cached_properties = set(self._get_cached_properties())

    if names:
      for name in names:
        if name in cached_properties:
          setattr(self, name, _CacheRefresh(expiration))
        else:
          raise AttributeError('No cached property %r on %r.' % (name, self))
    else:
      for varname in cached_properties:
        setattr(self, varname, _CacheRefresh(expiration))

    if remove_deleted:
      for varname in self._cache:
        if not varname in cached_properties:
          del self._cache[varname]

    try:
      self._cache.changed()
    except:
      pass

  def view_cache(self):
    """Get the age of cached values.

    :rtype: dict
    
    Properties not yet cached will appear as ``None``.
    
    """
    rv = dict.fromkeys(self._get_cached_properties(), None)
    if self._cache:
      now = time()
      rv.update(dict(
        (k, now - v[1])
        for k, v in self._cache.items()
      ))
    return rv

  @classmethod
  def cached_property(cls, func):
    """Decorator that turns a class method into a cached property.

    :param func: bound method to be turned in to a property
    :type func: func

    A cached property acts similarly to a property but is only computed once and
    then stored in the instance's ``_cache`` attribute along with the time it was
    last computed. Subsequent calls will read directly from the cached value.  To
    refresh several or all cached properties, use the ``refresh_cache`` method.

    Should only be used with methods of classes that inherit from ``Cacheable``.
    
    """
    return _CachedProperty(func)


class _CacheRefresh(object):

  """Special class used to trigger cache refreshes."""

  def __init__(self, expiration):
    self.expiration = expiration


class _CachedProperty(property):

  """Instance of a cached property for a model.

  Based on the emulation of PyProperty_Type() in Objects/descrobject.c from 
  http://infinitesque.net/articles/2005/enhancing%20Python's%20property.xhtml

  """

  def __init__(self, func):
    self.func = func
    self.__doc__ = func.__doc__

  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    else:
      try:
        return obj._cache[self.func.__name__][0]
      except (KeyError, TypeError) as e:
        value = self.func(obj)
        self.__set__(obj, value)
        return value

  def __set__(self, obj, value):
    if not obj._cache:
      obj._cache = {}
    if value:
      if isinstance(value, _CacheRefresh):
        if self.func.__name__ in obj._cache:
          t = time() - obj._cache[self.func.__name__][1]
          if t > value.expiration:
            obj._cache[self.func.__name__] = (self.func(obj), time())
        else:
          obj._cache[self.func.__name__] = (self.func(obj), time())
      else:
        obj._cache[self.func.__name__] = (value, time())
      try:
        obj._cache.changed()
      except:
        pass

  def __repr__(self):
    return '<CachedProperty %r>' % self.func




#!/usr/bin/env python

"""General helpers."""

from __future__ import absolute_import

from collections import Mapping, namedtuple
from ConfigParser import SafeConfigParser
from copy import deepcopy
from csv import DictReader
from datetime import datetime
from decimal import Decimal
from json import loads
from re import sub
from sqlalchemy.orm import Query


def convert(value, rtype=None, allow_json=False):
  """Converts a string to another value.

  :param value: the value to be converted
  :type value: str
  :param rtype: string representation of the type the value should be
    converted to. Accepted values are ``['int', 'float', 'bool', 'unicode',
    'str', 'json']``.
  :type rtype: str
  :param allow_json: allow loading of json strings
  :type allow_json: bool
  :rtype: int, float, bool, str, unicode

  If ``rtype`` isn't specified, the following conversions are attempted in
  order: ``int``, ``float``, ``bool`` and finally ``json.loads`` (only if
  ``allow_json`` is ``True``). If all fail, the method returns the
  string ``value`` unchanged.

  Boolean conversions convert ``'true', '1'`` to ``True``, and ``'false', '0'``
  to ``False`` (case insensitive) and otherwise raises an error.

  """
  value = value.strip()
  if rtype:
    if rtype == 'int':
      return int(value)
    elif rtype == 'float':
      return float(value)
    elif rtype == 'bool':
      if value.lower() == 'true' or value == '1':
        return True
      elif not value or value.lower() == 'false' or value == '0':
        return False
      else:
        raise ValueError('Can\'t convert %s to boolean.' % value)
    elif rtype == 'unicode':
      return unicode(value, encoding='utf-8', errors='replace')
    elif rtype == 'str':
      return value
    elif rtype == 'json':
      return loads(value)
    # if we get here, something has gone wrong
    raise ValueError('Invalid conversion type: %s' % rtype)
  else:
    try:
      return int(value)
    except ValueError:
      try:
        return float(value)
      except ValueError:
        if value.lower() == 'true':
          return True
        elif value.lower() == 'false':
          return False
        elif value == 'None':
          return None
        else:
          if allow_json:
            try:
              return loads(value)
            except ValueError:
              pass
          return value

def uncamelcase(name):
  """Transforms CamelCase to underscore_case.

  :param name: string input
  :type name: str
  :rtype: str
  
  """
  s1 = sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
  return sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def to_json(value, depth=1):
  """Serialize an object.

  :param value: the object to be serialized.
  :type value: varies
  :param depth: used when serializing nested objects with a ``to_json``
    method. In that case, the ``depth`` parameter is decremented by one each
    call. This paramater sets the initial value.
  :type depth: int
  :rtype: varies

  """
  if hasattr(value, 'to_json'):
    return value.to_json(depth - 1)
  if isinstance(value, dict):
    return {k: to_json(v, depth) for k, v in value.items()}
  if isinstance(value, list):
    return [to_json(v, depth) for v in value]
  if isinstance(value, (float, int, long, str, unicode, tuple)):
    return value
  if value is None:
    return None
  if isinstance(value, datetime):
    return str(value)
  if isinstance(value, Decimal):
    return float(value)
  raise ValueError('Not jsonifiable')

def parse_config(file_or_filepath, default=None, allow_json=False,
  case_sensitive=False, parser_type=SafeConfigParser):
  """Returns a dictionary of values from a configuration file.

  :param file_or_filepath: file or filepath to configuration file
  :type file_or_filepath: str or file
  :param default: dictionary of default values to use
  :type default: dict
  :param allow_json: allow loading of json options
  :type allow_json: bool
  :param case_sensitive: keep option names' case
  :type case_sensitive: bool
  :param parser_type: base parser type to use for parsing the file
  :type parser_type: ConfigParser.RawConfigParser
  :rtype: dict

  """
  parser = parser_type()
  if case_sensitive:
    parser.optionxform = str
  if isinstance(file_or_filepath, str):
    with open(file_or_filepath) as f:
      parser.readfp(f)
  else:
    parser.readfp(file_or_filepath)
  conf = {
    s: {
      k: convert(v, allow_json=allow_json)
      for (k, v) in parser.items(s)
    }
    for s in parser.sections()
  }

  if default:
    # nested dictionary update function
    def update(a, b):
      for k, v in b.iteritems():
        if isinstance(v, Mapping):
          a[k] = update(a.get(k, {}), v)
        else:
          a[k] = b[k]
      return a

    conf = update(deepcopy(default), conf)

  return conf

Part = namedtuple('Part', ['offset', 'limit'])

def partition(collection, parts=0, size=0):
  """Split an iterable into several pieces.

  :param collection: the iterable that will be partitioned. Note that
    ``partition`` needs to be able to compute the total length of the iterable
    so generators won't work.
  :type collection: list, query or file
  :param parts: number of parts to split the collection into
  :type parts: int
  :param size: number of items (lines if ``collection`` is a file) per
    part
  :type size: int
  :rtype: generator

  Only one of ``parts`` and ``size`` can be specified at a time.

  ``partition`` returns a generator that yields a tuple ``(batch, part)``
  on each iteration. ``batch`` is of the same type as ``collection``, filtered
  to the corresponding partition and ``part`` is a named tuple with two 
  properties:

  * ``offset``, the first index of the partition (if ``collection`` is a file
    it will be the first line number instead)
  * ``limit``, the max-length of the partition (the last one might be shorter)

  """
  if (parts and size) or (not parts and not size):
    raise ValueError('Exactly one of parts and size must be specified.')
  offset = 0
  limit = size
  if isinstance(collection, Query):
    # collection is a SQLAlchemy query
    # count won't be very efficient, but only run once so it's ok
    total = collection.count()
    if parts:
      limit = int(ceil(float(total) / parts))
    while offset < total:
      yield collection.offset(offset).limit(limit), Part(offset, limit)
      offset += limit
  elif isinstance(collection, list):
    total = len(collection)
    if parts:
      limit = int(ceil(float(total) / parts))
    while offset < total:
      yield collection[offset:offset + limit], Part(offset, limit)
      offset += limit
  elif isinstance(collection, file):
    total = sum(1 for line in collection)
    if parts:
      limit = int(ceil(float(total) / parts))
    while offset < total:
      collection.seek(0)
      yield islice(collection, offset, offset+limit), Part(offset, limit)
      offset += limit

def prod(iterable, key=None):
  """Cumulative product function (the equivalent of ``sum``).

  :param key: function called on each element of the iterable, if none then
    identity is assumed
  :type key: callable
  :rtype: int, float

  """
  rv = 1
  for elem in iterable:
    if key is None:
      rv *= elem
    else:
      rv *= key(elem)
  return rv


class SmartDictReader(DictReader):

  """``DictReader`` with built-in value conversion.

  :param csvfile: open file instance.
  :type csvfile: file
  :param fields: list of ``fieldnames`` or list of tuples
    ``(fieldname, fieldtype)``. If specified, the ``fieldtype`` will be passed
    as second argument to the ``convert`` function.
  :type fields: list
  :param silent: whether or not to silence errors while processing the file.
  :type silent: bool
  :param allow_json: allow loading of json strings
  :type allow_json: bool
  :param kwargs: keyword arguments to forward to the undelying
    ``csv.DictReader`` object.
  :rtype: iterable

  Interesting values for kwargs can be:

  * delimiter = '\t'
  * quotechar = '\x07'

  The following attributes are also available:

  * ``rows_imported``, the total number of rows successfully imported
  * ``errors``, a list of tuples ``(e, row)`` where ``e`` is error and ``row``
    the full row for each error raised

  """

  def __init__(self, csvfile, fields=None, silent=False, allow_json=False,
               **kwargs):
    self.csvfile = csvfile
    self.rows_imported = 0
    self.errors = []
    self.silent = silent
    self.allow_json = allow_json
    if fields:
      if isinstance(fields[0], (list, tuple)):
        kwargs['fieldnames'] = [field[0] for field in fields]
        self.field_types = dict(fields)
      else:
        kwargs['fieldnames'] = fields
        self.field_types = dict.fromkeys(fields, None)
      DictReader.__init__(self, csvfile, **kwargs)
    else:
      DictReader.__init__(self, csvfile, **kwargs)
      self.field_types = dict.fromkeys(self.fieldnames, None)

  def next(self):
    row = DictReader.next(self)
    try:
      processed_row = dict(
          (key, convert(value, self.field_types[key], self.allow_json))
          for key, value in row.iteritems()
      )
    except ValueError as e:
      self.errors.append((e, row))
      if not self.silent:
        raise e
    else:
      self.rows_imported += 1
      return processed_row

